@@ -1,7 +1,7 @@
 package QWizard::Generator::Tk;
 
 use strict;
-my $VERSION = '2.0.1';
+my $VERSION = '2.2.1';
 use Tk;
 use Tk::Table;
 use Tk::Pane;
@@ -79,7 +79,7 @@ sub new {
     $self->add_handler('graph',
 		       \&QWizard::Generator::Tk::do_graph,
 		       [['norecurse','values'],
-			['norecursemulti','image_options']]);
+			['norecursemulti','graph_options']]);
     $self->add_handler('image',
 		       \&QWizard::Generator::Tk::do_image,
 		       [['norecurse','imgdata'],
@@ -99,8 +99,9 @@ sub new {
 
 sub goto_top {
     my $self = shift;
+    my $wiz = shift;
     $self->unmake_top();
-    $self->clear_params();
+    $wiz->reset_qwizard();
 }
 
 sub goto_next {
@@ -198,20 +199,22 @@ sub init_screen {
 }
 
 sub do_ok_cancel {
-  my ($self, $nexttext, $dontdocan) = @_;
+  my ($self, $nexttext, $wiz, $dontdocan) = @_;
   if (!$self->{'bot'}) {
       $self->{'bot'} = $self->{'top'}->Frame(-relief => 'raised',
 					     -borderwidth => 3);
       if (!$self->{'prevbut'}) {
 	  $self->{'prevbut'} = 
-	    $self->{'bot'}->Button(-text => ('Back'),
+	    $self->{'bot'}->Button(-text => ($wiz->{'back_text'} || 
+						 'Back'),
 				   -command => [\&goto_prev, 
 						$self]);
 	  $self->{'prevbut'}->pack(-side => 'left');
       }
       if (!$self->{'nextbut'}) {
 	  $self->{'nextbut'} = 
-	    $self->{'bot'}->Button(-text => ($nexttext || 'Ok'),
+	    $self->{'bot'}->Button(-text => ($nexttext  ||
+					     $wiz->{'next_text'} || 'Next'),
 				   -command => [\&goto_next, 
 						$self]);
 	  $self->{'nextbut'}->pack(-side => 'left');
@@ -219,8 +222,9 @@ sub do_ok_cancel {
       if (! $dontdocan) {
 	  if (!$self->{'canbut'}) {
 	      $self->{'canbut'} = 
-		$self->{'bot'}->Button(-text => 'Cancel',
-				       -command => [\&goto_top, $self]);
+		$self->{'bot'}->Button(-text => ($wiz->{'cancel_text'} || 
+						 'Cancel'),
+				       -command => [\&goto_top, $self, $wiz]);
 	      $self->{'canbut'}->pack(-side => 'right');
 	  }
       }
@@ -287,7 +291,7 @@ sub set_default {
 
 sub wait_for {
   my ($self, $wiz, $next) = @_;
-  $self->do_ok_cancel($next);
+  $self->do_ok_cancel($next, $wiz);
   $self->our_mainloop();
   return 1;
 }
@@ -305,19 +309,29 @@ sub do_error {
 sub do_question {
     my ($self, $q, $wiz, $p, $text, $qcount) = @_;
     my $top = $self->{'qtable'};
+
     $self->{'currentq'} = $qcount + $self->{'qadd'};
-    if ($q->{'helpdesc'} && !$self->qwpref('usehelpballons')) {
+
+    #
+    # Get the actual help text, in case this is a subroutine.
+    #
+    my $helptext = $q->{'helpdesc'};
+    if (ref($helptext) eq "CODE") {
+	$helptext = $helptext->();
+    }
+
+    if ($helptext && !$self->qwpref('usehelpballons')) {
 	my $f = $top->Frame();
 	$f->Label(-text => $text, -anchor => 'nw')->pack(-anchor => 'w');
-	$f->Label(-text => $q->{'helpdesc'}, -anchor => 'nw',
+	$f->Label(-text => $helptext, -anchor => 'nw',
 		  -font => 'Helvetica 12 italic')
 	  ->pack(-anchor => 'w');
 	$self->put_it($f, undef, 1);
     } else {
 	my $l = $top->Label(-text => $text, -anchor => 'nw');
 	$self->put_it($l, undef, 1);
-	if ($self->{'balloon'} && $q->{'helpdesc'}) {
-	    $self->{'balloon'}->attach($l, -balloonmsg => $q->{'helpdesc'});
+	if ($self->{'balloon'} && $helptext) {
+	    $self->{'balloon'}->attach($l, -balloonmsg => $helptext);
 	    # XXX: change the "help" window text, which doesn't exist yet.
 	}
     }
@@ -755,7 +769,9 @@ sub open_branch {
 	return;
     }
 
-    foreach my $child (@{$q->{'children'}->($wiz, $tree->infoData($branch))}) {
+    my $children = $q->{'children'}->($wiz, $tree->infoData($branch));
+    return if (!$children || $#$children == -1);
+    foreach my $child (@$children) {
 	add_node($wiz, $tree, $child, $q, $branch, $labels, @expand);
     }
 }
@@ -806,7 +822,7 @@ sub end_confirm {
     my ($self, $wiz) = @_;
     # this will be deleted by the cancel button if they press it.
     $self->do_hidden($wiz, 'wiz_confirmed', 'Commit');
-    $self->do_ok_cancel("Commit");
+    $self->do_ok_cancel("Commit", $wiz);
     $self->our_mainloop();
     return 1;
 }
@@ -838,7 +854,7 @@ sub start_actions {
 sub end_actions {
     my ($self, $wiz) = @_;
     $self->put_it('Done',3,1);
-    $self->do_ok_cancel("Finish");
+    $self->do_ok_cancel("Finish", $wiz);
     $self->clear_params();
     $self->our_mainloop();
     return 1;
