@@ -1,6 +1,6 @@
 package QWizard;
 
-our $VERSION = '2.0.1';
+our $VERSION = '2.2.1';
 require Exporter;
 
 our @ISA = qw(Exporter);
@@ -15,10 +15,12 @@ our $qwdebug_indent = 0;
 our $qwvar;
 our $qwcurrent;
 
-our %states = { ASKING => 1,
+our %states = ( ASKING => 1,
 		CONFIRMING => 2,
-		ACTING => 3
-	      };
+		ACTING => 3,
+		FINISHED => 4,
+		CANCELED => 5,
+	      );
 
 use strict;
 
@@ -50,6 +52,44 @@ sub new {
     return $self;
 }
 
+#
+# $wiz->run_hooks(NAME, HOOK_ARGS)
+#   runs all hooks bound to NAME with a passed reference to QWIZARD,
+#   earlier passed MAGIC_ARGS (see below) and a copied ARGS in an array 
+#   reference, which is passed in from the calling args.  The result is:
+#
+#     CODE->($wiz, MAGIC_ARGS, [HOOK_ARGS])
+#
+sub run_hooks {
+    my $self = shift;
+    my $hookname = shift;
+    my @args = @_;
+    if (exists($self->{'hooks'}{$hookname})) {
+	foreach my $hook (@{$self->{'hooks'}{$hookname}}) {
+	    $hook->{'code'}($self, @{$hook->{'args'}}, \@args);
+	}
+    }
+}
+
+#
+# $wiz->add_hooks(NAME, SUB_REF, MAGIC_ARGS)
+#   Adds a SUB_REF code block to the NAME set of hooks.  Optionally, a
+#   set of MAGIC_ARGS may be passed as well.  For portability ease, it is
+#   suggested that MAGIC_ARGS be a single argument of an array reference
+#   of any arguments that need passing.
+#
+sub add_hook {
+    my $self = shift;
+    my $hookname = shift;
+    my $coderef = shift;
+    my @magic_args = @_;
+    my $hook_definition = {'code' => $coderef};
+    push @{$self->{'hooks'}{$hookname}}, $hook_definition;
+    if ($#magic_args > -1) {
+	$hook_definition->{'args'} = \@magic_args;
+    }
+}
+
 ########################################################################
 # parseprefs()
 #
@@ -171,6 +211,7 @@ sub magic {
 
   $self->{'last_screen'} = 0;
 
+  $self->run_hooks('start_magic');
   do {
       qwdebug("------------------------------------------------------------------");
       qwdebug("incoming: " . ref($self->{'generator'}) . " :" 
@@ -190,6 +231,7 @@ sub magic {
       }
   } while ((!$self->{'last_screen'} && !$self->{'one_pass'} && 
 	    !$self->{'generator'}{'one_pass'}));
+  $self->run_hooks('end_magic');
 }
 
 sub do_named_primaries {
@@ -275,8 +317,8 @@ sub keep_working {
 			     my $num = $_[1];
 			     if ($_[0]{'done'} == 2) {
 				 $$num++;
-				 $_[2]->munge_form_data($_[2]->get_primary($_[0]{'name'}));
 				 qwdebug("munging form data for primary $_[0]->{name}");
+				 $_[2]->munge_form_data($_[2]->get_primary($_[0]{'name'}));
                              }
                          }, \$num, $self);
   qwdebug("$num primaries ran last time\n");
@@ -296,9 +338,11 @@ sub keep_working {
 			     sub {
 				 $_[0]{'done'} = 0 if ($_[0]{'done'} == 2);
 			     });
+      qwparam('redo_screen',0);
   } else {
       if ($num > 0) {
 	  # run post_answers clauses and mark as done
+	  $self->run_hooks('post_answers_begin');
 	  qwdebug("Running post answers clauses");
 	  $self->foreach_primary($self->{'active'},
 		sub {
@@ -328,6 +372,7 @@ sub keep_working {
 		        $_[0]{'done'} = 1;
 
 		    }}, $self);
+	  $self->run_hooks('post_answers_end');
       }
   }
 
@@ -339,9 +384,10 @@ sub keep_working {
       qwdebug("no screens to display, entering confirmation mode");
       if ($self->qwparam('no_actions') || !$self->has_actions() ||
 	  !$self->confirm_or_run_actions()) {
+
 	  # cancel button hit, restart at top
-	  $self->do_named_primaries(@otherwise);
-	  $self->{'state'} = $states{'ASKING'};
+	  $self->reset_qwizard();
+	  $self->{'state'} = $states{'CANCELED'};
       }
   } else {
       $self->{'state'} = $states{'ASKING'};
@@ -472,6 +518,7 @@ sub add_todos {
 	$order = $arg if ($arg eq '-early');
 	$order = $arg if ($arg eq '-late');
 	$merge = 1 if ($arg eq '-merge');
+	$count = $merge = shift if ($arg eq '-mergen');
 	$actionslast = 1 if ($arg eq '-actionslast');
 	$remap .= shift if ($arg eq '-remap');
     }
@@ -493,7 +540,7 @@ sub add_todos {
 	    push @{$addto->{'children'}}, $newchild;
 	}
     }
-    if ($#_ > -1 && $merge) {
+    if (($#_ > -1 || $merge > 1) && $merge) {
 	if ($order eq '-early') {
 	    $addto->{'children'}[0]{'merge'} = $count;
 	} else {
@@ -502,23 +549,6 @@ sub add_todos {
     }
 }
 
-sub qw_display_table {
-    my $it = shift;
-    my $self;
-print STDERR "QW:qw_display_table:  down in\n";
-    if (ref($it) eq "QWizard") {
-	$self = $it;
-	$it = shift;
-    } else {
-	$self = $qwcurrent;
-    }
-    if (!$self->{'generator'}) {
-	print STDERR "ack: no generator\n";
-    }
-    return $self->{'generator'}->do_table($self, $it, @_);
-}
-
-
 # Takes a list of "stuff" and executes it.  code snippets returns
 # errors, and strings may be include directly in the list which will
 # be returned as comments/output to the user
@@ -585,7 +615,7 @@ sub do_primary_action_list {
 		if (ref($desc) eq 'CODE') {
 		    $desc = $desc->($self);
 		}
-		if ($desc) {
+		if ($desc && $desc ne 'OK') {
 		    $self->{'generator'}->do_confirm_message($self,
 							     $self->unparamstr($desc));
 		}
@@ -595,6 +625,7 @@ sub do_primary_action_list {
 	if (exists($p->{'actions'})) {
 	    $results = $self->do_list(@{$p->{'actions'}});
 	    foreach my $r (@$results) {
+		next if ($r eq 'OK');
 		if ($r =~ /Error: (.*)/) {
 		    # XXX: revert entire DB transaction.
 		    qwdebug("  ERROR: $1");
@@ -669,9 +700,15 @@ sub confirm_or_run_actions {
 	    qwdebug("found end_section_hook. running.");
 	    $self->{'end_section_hook'}($self);
 	}
-	$self->{'last_screen'} = 1;
+
+	# Reset and start over.
+	$self->reset_qwizard();
     } elsif ($self->qwparam('wiz_canceled')) {
 	qwdebug("user hit the cancel button");
+
+	# Reset and start over.
+	$self->reset_qwizard();
+	$self->{'state'} = $states{'CANCELED'};
 	return 0;
     } else {
 
@@ -693,6 +730,13 @@ sub confirm_or_run_actions {
     return 1;
 }
 
+sub reset_qwizard {
+    my $self = shift;
+    $self->{'last_screen'} = 1;
+    $self->{'generator'}->clear_params();
+    $self->{'state'} = $states{'FINISHED'};
+}
+
 sub save_queue {
     my ($self, $queue, $name) = @_;
     $name = "wizard_queue_list" if (!$name);
@@ -930,9 +974,11 @@ sub ask_questions {
 	if ($p->{'questions'}) {
 	  $self->{'qcount'} = -1;
 	  qwdebug("starting queston processing/display");
+	  $self->run_hooks('ask_questions_begin');
 	  $self->{'generator'}->start_questions($self, $p,
 						$self->unparamstr($self->get_value($p->{title})),
 						(qwpref('pref_intro') ne '0' ? $self->unparamstr($self->get_value($p->{introduction})) : ""));
+	  $self->run_hooks('ask_questions_started');
 	  my @questions = @{$p->{'questions'}};
 	  my ($q, @newqs);
 
@@ -957,13 +1003,12 @@ sub ask_questions {
 	      delete $pdesc->{'error'};
 	  }
 	  foreach $q (@newqs) {
-	      $self->{'qcount'}++;
-	      $self->{'allqcount'}++;
 	      $self->{'currentq'} = $q; # for generators mostly
 	      $self->{'currentp'} = $p; # for generators mostly
 	      push @passvars, $self->ask_question($p, $q);
 	  }
 	  $self->{'generator'}->end_questions($self, $p);
+	  $self->run_hooks('ask_questions_end');
         }
 
 	if ($p->{'sub_modules'}) {
@@ -987,6 +1032,7 @@ sub ask_questions {
 	}
 	$repeat--;
 	$pdesc->{'done'} = 2;
+
 	if ($repeat <= 0) {
 	    qwdebug("Done processing screen's primaries ");
 	    if ($self->{'allqcount'}) {
@@ -1002,14 +1048,19 @@ sub ask_questions {
 
 		# call the generator's wait_for routine to wrap up.
 		my $next = (qwparam('no_actions') ne '' || !$self->has_actions_or_post_answers()) ? 'Finished' : 'Next';
+		if (qwparam('no_actions') ne '' || !$self->has_actions_or_post_answers()) {
+		    $self->{'last_screen'} = 1;
+		}
 		$next = qwparam('QWizard_next') if (qwparam('QWizard_next'));
 		if ($self->{'generator'}->wait_for($self, $next, $p)) {
 		    if (ref($self->{'end_section_hook'}) eq "CODE") {
 			qwdebug("end_section_hook found.  Calling.\n");
 			$self->{'end_section_hook'}($self, $p);
 		    }
+		    $self->reset_qwizard() if ($self->{'last_screen'});
 		    return 0;
 		}
+		$self->reset_QWizard() if ($self->{'last_screen'});
 	    } else {
 		# no questions generated by current screen, try one more.
 		$repeat++;
@@ -1024,6 +1075,10 @@ sub ask_question {
 
     my $pdesc = $self->{'context'};
 
+    # increment the question counts
+    $self->{'qcount'}++;
+    $self->{'allqcount'}++;
+
     if (ref($q) ne "HASH") {
 	$self->{'generator'}->do_separator($q, $self, $p, $q);
     } else {
@@ -1046,8 +1101,9 @@ sub ask_question {
 	    $def = $self->qwparam($q->{'name'});
 	    qwdebug("  pulling original: $def\n");
 	} else {
-	    if (qwparam($q->{'name'}) && !$q->{'override'}) {
-		print STDERR "*** QWizard: Warning redefining value from question named $q->{name} (value was: " . qwparam($q->{'name'}) . ".  This may or may not have been intentional.  Add the 'override' flag to the second question, or rename the parent's question, to turn off this warning in the future\n";
+	    if (qwparam($q->{'name'}) &&
+		!$q->{'override'} && !$self->{'override'}) {
+		print STDERR "*** QWizard: Warning redefining value from question named $q->{name} (value was: " . qwparam($q->{'name'}) . ").  This may or may not have been intentional.  Add the 'override' flag to the second question, or rename the parent's question, to turn off this warning in the future\n";
 	    }
 	    $def = $self->get_value($q->{default});
 	}
@@ -1157,10 +1213,12 @@ sub start_page {
   my $self = shift;
   $qwcurrent = $self;
   qwdebug("starting page");
+  $self->run_hooks('start_page_begin');
   $self->{'generator'}->init_screen($self, 
 				    $self->{'generator'}->{'title'} ||
 				    $self->{'title'} ||
 				    'the wizard');
+  $self->run_hooks('start_page_end');
 }
 
 sub qwparam {
@@ -1619,6 +1677,10 @@ answers.
 
 =head1 SYNOPSIS
 
+  #
+  # The following code works as a application *or* as a CGI script both:
+  #
+
   use QWizard;
 
   my %primaries =
@@ -1648,6 +1710,11 @@ answers.
 
   $qw->magic('starting_node');
 
+
+  #
+  # PLEASE see the examples in the examples directory.
+  #
+
 =head1 DESCRIPTION
 
 QWizard displays a list of grouped questions, and retrieves and processes
@@ -1665,29 +1732,48 @@ the data has been gathered and verified, then it can be handled as appropriate
 (e.g., written to a database, used for system configuration, or used to
 generate a graph.)
 
-Current user interfaces that exist are HTML, Tk, and (minimally) ReadLine.
-A single QWizard script implementation can make use of any of the output
-formats without code modification.  Thus it is extremely easy to write
-portable I<wizard> scripts that can be used without modification by both
-graphical window environments (Tk) and HTML-based web environments (e.g., CGI
-scripts.)
+Current user interfaces that exist are HTML, Gtk2, Tk, and (minimally)
+ReadLine.  A single QWizard script implementation can make use of any
+of the output formats without code modification.  Thus it is extremely
+easy to write portable I<wizard> scripts that can be used without
+modification by both graphical window environments (Gtk2 and Tk) and
+HTML-based web environments (e.g., CGI scripts.), as well with
+intercative command line enviornments (ReadLine).
 
-Back-end interfaces (child classes of the I<QWizard::Generator> module) are
-responsible for displaying the information to the user.  Currently HTML and
-Tk, in that order, are the two output mechanisms that work the best.  Others
-are planned, but are not far along in development.  Developing new generator
-back-ends is fairly simple and doesn't take a lot of code (assuming the
-graphic interface is fairly powerful and contains a widget library.)
+Back-end interfaces (child classes of the I<QWizard::Generator>
+module) are responsible for displaying the information to the user.
+Currently HTML, Gtk2, Tk and ReadLine, are the output mechanisms that
+work the best (in that order).  Some others are planned (namely a
+curses version), but are not far along in development.  Developing new
+generator back-ends is fairly simple and doesn't take a lot of code
+(assuming the graphic interface is fairly powerful and contains a
+widget library.)
 
 QWizard operates by displaying a series of "screens" to the user.  Each screen
 is defined in a QWizard construct called a I<primary> that describes the
 attributes of a given screen, including the list of I<questions> to be
 presented to the user.  Primaries can contain questions, things to do
 immediately after the questions are answered (I<post_answers>), and things
-to do once the entire series of screens have been answered (I<actions>.)
+to do once the entire series of screens have been answered (I<actions>).
 Other information, such as a title and an introduction, can also be attached
 to a primary.
 
+An example very minimal primary definition containing one question:
+
+  my %primaries = (
+    myprimary =>
+    {
+      title => "my screen title",
+      introduction => "optional introduction to the screen",
+      questions =>
+      [
+        {
+          type => 'checkbox',
+          text => 'Should the chicken cross the road?',
+        }
+      ],
+    }
+
 After defining a set of primaries, a new QWizard object must be created.  The
 QWizard I<new>() constructor is given a set of options, such as window title
 and a reference to a hash table containing the primaries.  (The complete set
@@ -1695,14 +1781,20 @@ of options may be found in the "QWizard new() Options" section.) The question
 display and data collection is started by calling the I<magic>() routine of
 the new QWizard object.
 
-There are examples distributed with the QWizard module sources that may help
-to understand the whole system and what it is capable of.  See the B<examples>
-directory of the QWizard source code tree for details.  Also, QWizard was
-written mostly due to requirements of the Net-Policy project.  Net-Policy
-makes very extensive use of QWizard and is another good place to look for
-examples.  In fact, the QWizard CVS code is located inside the Net-Policy CVS
-tree.  See http://net-policy.sourceforge.net/ for details on the Net-Policy
-project.
+  my $qw = new QWizard(primaries => \%primaries,
+                       title => 'my title');
+  $qw->magic('myprimary');
+
+There are examples distributed with the QWizard module sources that
+may help to understand the whole system and what it is capable of.
+See the B<examples> directory of the QWizard source code tree for
+details.  Also, QWizard was written mostly due to requirements of the
+Net-Policy project.  Net-Policy makes very extensive use of QWizard
+and is another good place to look for examples.  In fact, the QWizard
+CVS code is located inside the Net-Policy CVS tree.  See
+http://net-policy.sourceforge.net/ for details on the Net-Policy
+project.  There are a number of screen shots showing all the
+interfaces as well on the main net-policy web site.
 
 =head2 MAGIC() PSEUDO-CODE
 
@@ -1759,17 +1851,18 @@ Current generator classes are:
 
   - QWizard::Generator::Best           (default: picks the best available)
   - QWizard::Generator::HTML
+  - QWizard::Generator::Gtk2
   - QWizard::Generator::Tk
   - QWizard::Generator::ReadLine       (limited in functionality)
 
 The I<QWizard::Generator::Best> generator is used if no specific generator
 is specified.  The I<Best> generator will create an HTML generator if used
 in a web context (i.e., a CGI script), or else pick the best of the available
-other generators (Tk, then ReadLine).
+other generators (Gtk2, then Tk, then ReadLine).
 
-This example forces a Tk generator to be used:
+This example forces a Gtk2 generator to be used:
 
-   my $wiz = new QWizard(generator => new QWizard::QWizard::Tk(),
+   my $wiz = new QWizard(generator => new QWizard::QWizard::Gtk2(),
                          # ...
                         );
 
@@ -1778,7 +1871,7 @@ This example forces a Tk generator to be used:
 This should be the top location of a web page where the questions will be
 displayed.  This is needed for "go to top" buttons and the like to work.  This
 is not needed if the QWizard-based script is not going to be used in a CGI
-or other web-based environment.
+or other web-based environment (eg, if it's going to be used in mod_perl).
 
 =item primaries => \%my_primaries
 
@@ -1938,10 +2031,11 @@ before the wizard is finished.
 # successful.  Returning anything else will print the result, as if it
 # is an error message, to the user.
 
-For HTML output, these will be run just before the next screen is printed
-after the user has submitted the answers back to the web server.  For
-window-based output (Tk, etc.) the results are similar and these subroutines
-are evaluated before the next window is drawn.
+For HTML output, these will be run just before the next screen is
+printed after the user has submitted the answers back to the web
+server.  For window-based output (Gtk2, Tk, etc.) the results are
+similar and these subroutines are evaluated before the next window is
+drawn.
 
 =item actions => [ VALUES ]
 
@@ -2454,11 +2548,6 @@ Adds the primary to the B<end> of the list of primaries to call, such
 that it is called last, unless another call to I<add_todos>() appends
 something even later.
 
-=item -merge
-
-Merges all the specified primaries listed into a single screen.  This has the
-effect of having multiple primaries displayed in one window.
-
 =item -actionslast
 
 All the actions of subsequent primaries that have been added as the result
@@ -2472,6 +2561,15 @@ own actions to be successful.
 However, this flag indicates that the actions of the childrens' primaries
 listed in this call are to be called B<before> the current primary's actions.
 
+=item -merge
+
+Merges all the specified primaries listed into a single screen.  This has the
+effect of having multiple primaries displayed in one window.
+
+Important note: you can not use both -remap (see below) and -merge at
+the same time!  This will break the remapping support and you will not
+get expected results!
+
 =item -remap => 'NAME'
 
 If a series of questions must be called repeatedly, you can use this flag to
@@ -2487,6 +2585,10 @@ is an example that illustrates this in the QWizard Perl module source code
 B<examples> directory, in the file B<number_adding.pl>.  This code repeatedly
 asks for numbers from the user using the same primary.
 
+Important note: you can not use both -remap and -merge at the same
+time!  This will break the remapping support and you will not get
+expected results!
+
 =back
 
 =item $qw->add_primary(key => value, key => value);
@@ -2565,10 +2667,11 @@ following:
 
 =head1 EXAMPLES
 
-There are a few usage examples in the B<examples> directory of the source
-package.  These examples can be run from the command line or installed as a
-CGI script without modification.  They will run as a CGI script if run from
-a web server, or will launch a Tk window if run from the command line.
+There are a few usage examples in the B<examples> directory of the
+source package.  These examples can be run from the command line or
+installed as a CGI script without modification.  They will run as a
+CGI script if run from a web server, or will launch a Gtk2 or Tk
+window if run from the command line.
 
 =head1 EXPORT
 
